(* $Id: pr_m4.ml 1440 2006-06-21 13:35:02Z roglo $ *)

open Printf

type m4 = Pa_m4.m4 =
    Macro of MLast.loc * string * (MLast.loc * (string * string)) list
  | Comment of MLast.loc * string
  | QuotedString of MLast.loc * string
  | LeadingSpace of MLast.loc * string

let not_impl loc name x =
  let desc =
    if Obj.tag (Obj.repr x) = Obj.tag (Obj.repr "") then
      sprintf "\"%s\"" (Obj.magic x)
    else if Obj.is_block (Obj.repr x) then
      "tag = " ^ string_of_int (Obj.tag (Obj.repr x))
    else "int_val = " ^ string_of_int (Obj.magic x)
  in
  Stdpp.raise_with_loc loc
    (Failure (sprintf "mm4: not impl: %s; %s" name desc))

let eprint_location loc fn =
  let (_, line, bp, ep) = Stdpp.line_of_loc "" loc in
  eprintf "File \"%s\", line %d, characters %d-%d:\n" fn line bp ep

let path = ref []
let path_open_in x =
  let rec loop =
    function
      [] -> open_in x
    | d :: dl ->
        try open_in (Filename.concat d x) with Sys_error _ -> loop dl
  in
  loop !path

exception Unbound_macro

let macros_tab = Hashtbl.create 1003

let macro_include loc fn s sl =
  let fn =
    match sl with
      [(_, (_, s))] -> s
    | _ -> raise Unbound_macro
  in
  let v = !Pcaml.input_file in
  Pcaml.input_file := fn;
  begin try
    let ic = path_open_in fn in
    let _ = Grammar.Entry.parse Pcaml.implem (Stream.of_channel ic) in
    close_in ic;
    Pcaml.input_file := v
  with
    exc ->
      Pcaml.input_file := v;
      let exc =
        match exc with
          Stdpp.Exc_located (loc, exc) ->
            eprint_location loc fn;
            exc
        | exc -> exc
      in
      eprintf "mm4: %s\n" (Printexc.to_string exc);
      flush stderr
  end

let _ = Hashtbl.add macros_tab "include" macro_include

let apply_macro loc fn s sl =
  begin try
    let f =
      try Hashtbl.find macros_tab s with Not_found -> raise Unbound_macro
    in
    f loc fn s sl
  with
    Unbound_macro ->
      printf "%s" s;
      if sl <> [] then (
        printf "(";
        let _ =
          List.fold_left
            (fun sep (_, (ls, s)) -> printf "%s%s%s" sep ls s; ",") "" sl
        in
        printf ")";
        flush stdout
      )
      else ()
  end

let rec mm4 fn =
  function
    Macro (loc, s, sl) :: m4l -> apply_macro loc fn s sl; mm4 fn m4l
  | LeadingSpace (_, s) :: m4l -> printf "%s" s; flush stdout; mm4 fn m4l
  | Comment (_, s) :: m4l -> printf "#%s\n" s; flush stdout; mm4 fn m4l
  | x :: _ -> not_impl Token.dummy_loc "mm4" x
  | [] -> ()

let m4ltosil m4l = mm4 !Pcaml.input_file m4l; []

let add_in_path s = path := s :: !path

let _ = Pa_m4.m4ltosil := m4ltosil
let _ = Pcaml.print_implem := fun _ -> ()
let _ = Pcaml.add_option "-I" (Arg.String add_in_path) "Add in path"
